
import numpy as np

# Simulation defaults
_simulation = "statevector"
_max_bond_dim = 64

# Memory overheads
ram_overhead = 2
fastqsvm_overhead = 2

# DTypes (defaults)
state_dtype = np.complex128
kernel_dtype = np.float64